#
# ZAnnot.py - Zope Annotation Server
#
# Code by Yasushi Yamazaki and Brent Hendricks
#
# (C) 2002 Brent Hendricks - licensed under the terms of the
# GNU General Public License (GPL).  See LICENSE.txt for details

from OFS.Folder import Folder
from OFS.Image import File
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import getSecurityManager, ClassSecurityInfo
from random import randint
from binascii import a2b_hex
from urllib import quote
from DateTime import DateTime
from Products.ZSQLMethods.SQL import SQL
from AnnotRequestParser import AnnotRequestParser


### add interface ############################
manage_addZAnnotationForm = PageTemplateFile('zpt/manage_addZAnnotationForm',
                                            globals(),
                                            __name__='manage_addZAnnotationForm')


def manage_addZAnnotation(self, id, file='', title='', content_type='', REQUEST=None):
                        
    """Add a new ZAnnotation object.

    Creates a new ZAnnotation object 'id' with the contents of 'file'"""

    id=str(id)
    title=str(title)

    self=self.this()

    # First, we create the file without data:
    self._setObject(id, ZAnnotation(id, title))

    # Now we "upload" the data.  By doing this in two steps, we
    # can use a database trick to make the upload more efficient.
    if file:
        self._getOb(id).manage_upload(file)
    if content_type:
        self._getOb(id).content_type=content_type

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')


class ZAnnotation(File):
    """ZAnnotation"""

    meta_type = "Annotation"

    #manage_annotationForm = PageTemplateFile('zpt/manage_annotationForm',
    #                                         globals(),
    #                                         __name__='manage_annotationForm')

    manage_look = PageTemplateFile('zpt/manage_look',
                                   globals(),
                                   __name__='manage_look')

    #options = list(File.manage_options)
    #manage_options = tuple(options[:3]
    #                       + [{'label':'Annotation','action':'manage_annotationForm'}]
    #                       + options[3:])



    def __init__(self, id, title):
        return File.__init__(self, id, title,'','text/html')


    def annotationBody(self):
        """Return the annotation's body"""
        self.REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
        return str(self.data)


    def description(self):
        """Returns the RDF dexription of the annotation"""
        return self.annotationRDF(annotations=[self])

    # Default to the description
    index_html = description

    def manage_beforeDelete(self, item, container):
        # removal is not allowed if it has a reply
        uri = self.absolute_url()
        if self.searchChild({"annotates":uri,
                             "root":uri,
                             "inreplyto":uri}):
            raise "Forbidden", "The annotation with a reply cannot be removed."



manage_addZAnnotForm = PageTemplateFile("zpt/manage_addZAnnotForm",
                                        globals(),
                                        __name__="manage_addZAnnotForm")


def manage_addZAnnot(self,id,db=None,annotUsesDB=None,REQUEST=None):
    """Add a new ZAnnot object."""

    id = str(id)
    if id=='':
        raise "Bad Request", "Please specify an ID."

    # create
    self = self.this()
    #if annotUsesDB=="Yes":
    #    annotObj = ZAnnotDB(id, db)
    #else:
    annotObj = ZAnnot(id)
    self._setObject(id, annotObj)

    if REQUEST is not None:
        REQUEST["RESPONSE"].redirect(self.absolute_url()+"/manage_main")






### class definition ##########################
class ZAnnot(Folder):
    """Annotation Server Product."""
    
    # preferences
    meta_type = "Annotation Server"
    prefix = "annot"        # annotation object prefix (annot102034234ASFDFSDF, etc.)
    id_digits = 10

    # server-side management interface
    # it's probably for the best to not name this "index_html"
    main = PageTemplateFile("zpt/main", globals(), __name__="main")
    main.content_type = "text/html"
    main_name = "main"

    # Template for returning annotations
    annotationRDF = PageTemplateFile('zpt/annotation',globals(),__name__='annotation')
    annotationRDF.content_type="application/xml"
    annotationRDF_name = "annotation"

    # Template for annotation POST replies
    postreplyRDF = PageTemplateFile('zpt/postreply',globals(),__name__='postreply')
    postreplyRDF.content_type="application/xml"
    postreplyRDF_name = "postreply"


    def __init__(self,id):
        """Constructor."""
        self.id = id


    def index_html(self):
        """Handles client-issued requests.  All permission/error checking
           will be done in self.*Annotation subroutines."""
        # user request
        method = self.REQUEST['REQUEST_METHOD']
        # call appropriate routines
        if method=="GET" and self.REQUEST.has_key('w3c_annotates'):
            target = self.REQUEST['w3c_annotates']
            try:
                replyTree = self.REQUEST.form['w3c_replyTree']
            except KeyError:
                replyTree = None
            objs = self.queryAnnotation(target, replyTree)
            return self.annotationRDF(annotations = objs)
        elif method=="POST":
            # New annotation or update old one
            content = str(self.REQUEST['BODY'])
            # FIXME: we should make sure the body exists
            query_string = self.REQUEST['QUERY_STRING']
            if query_string:
                for query in query_string.split('&'):
                    (id, val) = query.split('=')
                    if id == "replace_source":
                        child_id = val.split('/')[-1]
                        return self.updateAnnotation(child_id, content)
            # If 'replace_source' isn't in the query string, just add the annotation
            return self.addAnnotation(content)
        # Update (PUT /annot/annotASDFASDF) - no client implements this yet
        #elif method="PUT"
        #    if child_id.startswith(self.prefix):
        #        return self.updateAnnotation(child_id, content)
        #else:
        #    return Folder.__getattr__('index_html')


    __call__ = index_html


    def main_execute(self,REQUEST):
        """Server-side management functions.  Should only be called from the management screen."""
        req = REQUEST
        # execute - only removal is implemented for now
        message = ''
        if req.has_key("remove"):
            # remove selected annotations
            if req.has_key("checked"):
                for annot_id in req.checked:
                    # let "Unauthorized" throw its exception
                    try:
                        # calling self.delAnnotation checks the user's edit permission first
                        self.delAnnotation(annot_id)
                        message = "Remove successful."
                    except "Not Found",e:
                        message = str(e)
                    except "Forbidden",e:
                        message = str(e)
            else:
                message = "Nothing was selected."
        # reconstruct the request vars (query/b_start needed for search result)
        q = ["message=%s" % quote(message)]
        if req.has_key("query"):
            q.append("query=%s" % quote(req.query))
        if req.has_key("b_start"):
            q.append("b_start=%s" % quote(req.b_start))
        req["RESPONSE"].redirect("%s?%s" % (self.main_name,'&'.join(q))) 

    def main_search(self,query):
        """Search function called by main.zpt"""
        # don't search for empty string
        query = query.strip()
        if query:
            time = DateTime().timeTime()
            search = self.searchChild({"annot_id":query,
                                       "annotates":query,
                                       "title":query,
                                       "creator":query,
                                       "root":query,
                                       "inreplyto":query,
                                       "personid":query})
            return ("%.3f" % (DateTime().timeTime()-time),search)
        return ('',[])
    

    # commands ------------------
    def addAnnotation(self, content):
        """Create an annotation."""
        self.setMessage("add request with %d bytes" % len(content))
        # edit permission check
        self.checkPermission()
        # parse RDF
        result =  self.createFromRDF(content)
        # Send out reply
        c = self.postreplyRDF(result=result)
        self.REQUEST.RESPONSE.setHeader("Location", result["location"])
        return c


    def updateAnnotation(self, child_id, content):
        """Update an existing annotation.
           First checks if the annotation exists & valid, and then checks the
           user's editing permission."""
        self.setMessage("update request for %s with %d bytes" % (child_id, len(content)))
        # edit permission check
        self.checkPermission(child_id)
        # parse RDF
        result = self.createFromRDF(content, id=child_id)
        # Send out reply
        c = self.postreplyRDF(result=result)
        self.REQUEST.RESPONSE.setHeader("Location", result["location"])
        return c


    def queryAnnotation(self, target, replyTree=None):
        """Search for annotations."""
        self.setMessage("query request")
        query = {"annotates":target}
        if replyTree:
            query['root'] = query['inreplyto'] = replyTree
        return self.searchChild(query)
        

    # subroutines --------------------
    def setMessage(self, content):
        """Send custom header for debugging."""
        self.REQUEST["RESPONSE"].setHeader("X-ZAnnot-Message", content)

    def quote(self,string):
        """urllib's quote function.  main.zpt calls this."""
        return quote(string)

    def checkPermission(self,child_id=''):
        """Checks the edit permission.
           It first checks if the user is authenticated,
           then it checks if the child exists & user has a write access to it.
           A user with a "Manager" role can edit any annotations.
           If no child_id is specified (new annotation case), it only checks for authentication."""
        # authenticate if needed
        roles = getSecurityManager().getUser().getRoles()
        if not "Manager" in roles and "Anonymous" in roles:
            # raising "Unauthorized" automatically sets WWW-Authenticate header.
            # to do it manually, set "WWW-Authenticate" to 'Basic realm="Zope"'
            raise "Unauthorized", "Please log in first."
        # if the child id is given, check permission on that child
        if child_id:
            # does the child exist?  this raises "Not Found" if not
            child = self.getChild(child_id)
            # is this an annotation object?  annotation body can't be uploaded directly
            if not hasattr(child,"personid"):
                raise "Forbidden", "You can only edit the annotation object."
            # check personid
            if not "Manager" in roles and self.getPersonID()!=child.personid:
                raise "Forbidden", "You have no permission to edit this annotation."

    def getPersonID(self):
        """Get the current user's zope personid."""
        personid = getSecurityManager().getUser()
        if hasattr(personid,"personid"):
            # extensible user folder case (depends on the system)
            personid = personid["personid"]
        return str(personid)

    def parsePath(self):
        """Determine the direct child id and grandchild id, if exists
           For example, http://server/my_id/some_id/some_id2/some_id3 would
             return (some_id, some_id2),
           where http://server/my_id/some_id would return (some_id, '').
           http://server/my_id and http://server/my_id/ are treated the same.
           """
        objs = self.REQUEST.get("PATH_INFO").split('/')
        child_id = ''
        child_function = ''
        # find my_id
        index = objs.index(self.id)
        try:
            child_id = objs[index+1]
            child_function = objs[index+2]
        except:
            pass
        return (child_id, child_function)

    def createFromRDF(self, content, id=''):
        """Parse the new/update request RDFs, then create the appropriate objects."""
        # parse
        try:
            p = AnnotRequestParser(content)
            r = p.parse()
        except Exception,e:
            raise "Bad Request", "Parsing Error (%s)" % str(e)
        # assign id if not given
        if not id:
            id = self.getNewID()
        # add annot_id to the resultset
        r.update({"annot_id": id})
        # annotation uri
        r.update({"location": "%s/%s" % (self.absolute_url(), id)})
        # create annotation and possibly a new body
        if not r["isBodyUri"]:
            # replace body content with body uri
            body = r["body"]
            r.update({"body": "/".join((self.absolute_url(), id, "annotationBody"))})
            # save annotation
        self.setAnnotChild(id, r, body)
        return r


    def getNewID(self):
        """Return a new ID."""
        # channel creation time
        id = str(self.REQUEST.environ["channel.creation_time"])
        # add a permutation of letters in ALL CAPS (ascii 65-90)
        id = id + a2b_hex(''.join([hex(randint(65,90))[2:] for i in range(self.id_digits)]))
        return self.prefix+id


    ####################################################################
    # Overridable functions - this implementation uses DTMLDocument for
    #   annotation storage, and File for annotation body storage.

    def setAnnotChild(self, id, props, body=None):
        """Create or update a child object.
           Expects a dictionary containing property-value pairs."""
        # for convenience
        p = ["annot_id","type","annotates","context","language","creator","created",
             "date","body","location","root","inreplyto"]

        if not hasattr(self, id):
            # new
            self.manage_addProduct["ZAnnot"].manage_addZAnnotation(id, file=body, title=props['title'], content_type='text/html')
            obj = getattr(self,id)
            # Create blank properties
            for name in p:
                obj._setProperty(name, props[name], "string")
            obj._setProperty("personid", self.getPersonID(), "string")
        else:
            # update
            obj = getattr(self,id)
            obj.title = props["title"]
            obj.update_data(body)
            for name in p:
                obj._updateProperty(name, props[name])


    def searchChild(self, map):
        """Search for the children with specified attributes.  All attributes are ORed."""
        searchString = " or ".join(["%s=='%s'" % (key,map[key]) for key in map.keys()])
        result = self.ZopeFind(self,None,None,None,searchString)
        return [obj[1] for obj in result]
    
    def getChild(self,id):
        """Get a child object."""
        if hasattr(self,id):
            return self[id]
        else:
            raise "Not Found", "The annotation \"%s\" does not exist." % id

        
    def delAnnotation(self, id):
        """Remove an annotation.
           First checks if the user has the editing permission."""
        self.setMessage("remove request for %s" % id)
        # edit permission check
        self.checkPermission(id)
        # remove the annotation
        self.manage_delObjects([id])
        # return 200
        return "Annotation Removed Successfully"

